/* com/syncbuilder/internal/sync/DlpImpl.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
* Copyright (C) 1998, Robert Crawford
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: DlpImpl.java,v 1.2 1998/12/09 09:50:36 christ Exp $ */

package com.syncbuilder.internal.sync;

import com.syncbuilder.internal.core.constants;

import com.syncbuilder.sync.DlpException;
import com.syncbuilder.storage.*;
import com.syncbuilder.device.*;

import com.syncbuilder.util.PackUtils;

/**
* This class implements the DLP in pure Java. It is here to
* avoid cluttering of the <code>com.syncbuilder.sync.Dlp</code> class, which 
* acts as a front-end to this class.
*
* @see com.syncbuilder.sync.Dlp
* @author Tilo Christ, Robert Crawford
*/
public final class DlpImpl
{
//============================================== Database ===================================================
	/**
	* Open a database on the Palm device through its cardnumber and its name, using the specified mode.
	*
	* @return the handle of the database
	*/
	public int openDB(String name, int mode, int cardno)
		throws DlpException
	{
		byte[] dlp_buf = new byte[name.length() + 3];
		dlp_buf[0] = (byte)cardno;
		dlp_buf[1] = (byte)mode;
		PackUtils.setZeroTermString(dlp_buf, 2, name); 

		byte[] result = dlp_exec(0x17, 0x20, dlp_buf);
  
	    _expect(result, 1);
  
	    int handle = result[0] & 0xff;

		return handle;
	}

	/**
	* Close a single open database.
	*/
	public void closeDB(int handle)
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];
		dlp_buf[0] = (byte)handle;
		
		dlp_exec(0x19, 0x20, dlp_buf);
	}


	public int createDB(String dbname, Char4 creator, Char4 type, int cardno, int flags, int version)
		throws DlpException
	{
		byte[] dlp_buf = new byte[46];

		PackUtils.setLong(dlp_buf, 0, creator.getInt());
		PackUtils.setLong(dlp_buf, 4, type.getInt());
		dlp_buf[8] = (byte)cardno;
		dlp_buf[9] = 0;
		PackUtils.setShort(dlp_buf, 10, flags);
		PackUtils.setShort(dlp_buf, 12, version);
		PackUtils.setZeroTermString(dlp_buf, 14, dbname);

		byte[] result = dlp_exec(0x18, 0x20, dlp_buf);

		_expect(result, 1);

		return (result[0] & 0xff);
	}


	/**
	* Get meta-information about a database on the Palm device.
	*
	* @return the meta-information about the database, or <code>null</code> if the database does not exist
	*/
	public DBInfo getDBInfo(int index, int cardno, int flags)
		throws DlpException
	{
		byte[] dlp_buf = new byte[4];
		dlp_buf[0] = (byte)flags;
		dlp_buf[1] = (byte)cardno;
		PackUtils.setShort(dlp_buf, 2, index);
		
		byte[] result = null;

		try
		{
			result = dlp_exec(0x16, 0x20, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.code == 5)
				return null;    // unfound database is indicated by null instead of an exception
			else
				throw de;
		}
		
		_expect(result, 48);
		
		DBInfo info = new DBInfo();
		
		info.more = (result[2] != 0); // More of what???

/*		if (pi_version(sd) > 0x0100) // PalmOS 2.0 has additional flag 
		    info->miscFlags = get_byte(result+5);
	    else
	    info->miscFlags = 0; 
		info.flagExcludeFromSync = ((dbflags & constants.dlpDBFlagExcludeFromSync) != 0); */

    	int dbflags = PackUtils.getShort(result, 6);
		info.flagReadOnly = ((dbflags & constants.dlpDBFlagReadOnly) != 0);
		info.flagResource = ((dbflags & constants.dlpDBFlagResource) != 0);
		info.flagOpen = ((dbflags & constants.dlpDBFlagOpen) != 0);
		info.flagBackup = ((dbflags & constants.dlpDBFlagBackup) != 0);
		info.flagAppInfoDirty = ((dbflags & constants.dlpDBFlagAppInfoDirty) != 0);
		info.flagNewer = ((dbflags & constants.dlpDBFlagNewer) != 0);
		info.flagReset = ((dbflags & constants.dlpDBFlagReset) != 0);

	    info.type = new Char4((int)(PackUtils.getLong(result, 8)));
    	info.creator = new Char4((int)(PackUtils.getLong(result, 12)));

	    info.version = PackUtils.getShort(result, 16);
	    info.modnum = PackUtils.getLong(result, 18);
	    info.createDate = getDate(result, 22);
	    info.modifyDate = getDate(result, 30);
	    info.backupDate = getDate(result, 38);
	    info.index = PackUtils.getShort(result, 46);

		info.name = PackUtils.getZeroTermString(result, 48, null);

		return info;
	}


	/**
	* Retrieve the number of records in the specified database.
	*/
	public int getRecordCount( int handle )
		throws DlpException
	{
		byte[] dlp_buf = new byte[1];
			
		dlp_buf[0] = (byte)handle;
	    byte[] result = dlp_exec(0x2B, 0x20, dlp_buf);

	    _expect(result, 2);
  
    	return PackUtils.getShort(result, 0);
	}


//===================================================== Record ===============================================

	/**
	* Retrieve a Record from the database. 
	*
	* @return <code>null</code> if the record could not be found, a specialization of <code>com.syncbuilder.storage.Record</code> otherwise.
	*/
	public Record readRecordByIndex(int handle, int index, DatabaseImpl dbclass)
		throws DlpException
	{
		byte[] dlp_buf = new byte[8];

		dlp_buf[0] = (byte)handle;
  		dlp_buf[1] = (byte)0x00;
  		PackUtils.setShort(dlp_buf, 2, index);
  		PackUtils.setShort(dlp_buf, 4, 0); // Offset into record 
  		PackUtils.setShort(dlp_buf, 6, DLP_BUF_SIZE); // length to return 

		byte[] result = null;

		try
		{
			result = dlp_exec(0x20, 0x21, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.code == 5)
				return null;    // unfound record is indicated by null instead of an exception
			else
				throw de;
		}

		_expect(result, 10);

    	int id = PackUtils.getLong(result, 0);
    	// getShort(result, 4) == index, but we already know the index
    	int size = PackUtils.getShort(result, 6);
    	int attr = result[8] & 0xff;
    	int category = result[9] & 0xff;

		byte[] content = new byte[result.length - 10];
    	System.arraycopy(result, 10, content, 0, content.length);

		return dbclass.createRecord(content, new RecordID(id), index, attr, category);
	}


	/**
	* Write the specified record to the specified database.
	* If the Record had a RecordID of 0 before, it will be
	* assigned a valid RecordID afterwards.
	*/
	public void writeRecord(int handle, Record record)
		throws DlpException
	{
		byte[] content = record.pack();

		byte[] dlp_buf = new byte[content.length + 8];

	    dlp_buf[0] = (byte)handle;
	  	dlp_buf[1] = (byte)0;
		PackUtils.setLong(dlp_buf, 2, record.getID().getValue());
	  	dlp_buf[6] = (byte)(record.getFlags());
	  	dlp_buf[7] = (byte)(record.getCategory());

	  	System.arraycopy(content, 0, dlp_buf, 8, content.length);

	  	byte[] result = dlp_exec(0x21, 0x20, dlp_buf);

	  	_expect(result, 4);

      	record.getID().setValue(PackUtils.getLong(result, 0)); /* New record ID */
	}


	/**
	* Delete a Record from the database. 
	*/
	public void deleteRecord(int handle, boolean all, RecordID id)
		throws DlpException
	{
		byte[] dlp_buf = new byte[6];

		dlp_buf[0] = (byte)handle;
		dlp_buf[1] = (byte) (all ? 0x80 : 0x00);
		PackUtils.setLong(dlp_buf, 2, id.getValue());

		byte[] result = null;

		try
		{
			result = dlp_exec(0x22, 0x20, dlp_buf);
		}
		catch(DlpException de)
		{
			if (de.code == 5)
				return;    // unfound record is silently ignored
			else
				throw de;
		}

		_expect(result, 0);

		return;
	}


//======================================== Palm device ======================================================

	/**
	* Read information about the HotSync user.
	*/	
	public UserInfo readUserInfo()
		throws DlpException
	{
    	byte[] result = dlp_exec(0x10, 0x00, null);

    	_expect(result, 30);

    	int user_len = result[28] & 0xff;

		UserInfo uinfo = new UserInfo();

    	uinfo.userID = PackUtils.getLong(result, 0);
    	uinfo.viewerID = PackUtils.getLong(result, 4);
    	uinfo.lastSyncPC = PackUtils.getLong(result, 8);
    	uinfo.successfulSyncDate = getDate(result, 12);
    	uinfo.lastSyncDate = getDate(result, 20);

		if (uinfo.lastSyncDate.getTime() < 18000)  // is lastSyncDate meaningful at all? And why 18000?
			uinfo.lastSyncDate = null;

    	int password_len = result[29] & 0xff;
    	uinfo.username = new String(result, 30, user_len);
		uinfo.password = new byte[password_len];
    	System.arraycopy(result, 30 + user_len, uinfo.password, 0, password_len);

    	return uinfo;
	}


	/**
	* Switch the message on the Palm device from
	* 'Identifying User' to 'Synchronizing'.
	*/
	public void switchMessage()
		throws DlpException
	{
		byte[] result = dlp_exec(0x2E, 0x00, null);

		_expect(result, 0);
	}

	
	public void addSyncLogEntry(String entry)
		throws DlpException
	{
		byte[] dlp_buf = new byte[entry.length() + 1];
		
		PackUtils.setZeroTermString(dlp_buf, 0, entry);
	
		dlp_exec(0x2A, 0x20, dlp_buf);
	}


	/**
	* End the HotSync procedure
	*/
	public void endHotSync(int status)
		throws java.io.IOException, DlpException
	{
		byte[] dlp_buf = new byte[2];
		PackUtils.setShort(dlp_buf, 0, status);

		byte[] result = dlp_exec(0x2F, 0x20, dlp_buf);
		
		_expect(result, 0);
	}


//============================================== Internal to DlpImpl =========================================

	/**
	* <b>!!!This constructor shall only be invoked by the Dlp class!!!</b>
	*/
	public DlpImpl( com.syncbuilder.sync.Socket socket )
		throws java.io.IOException
	{
		_socket = socket;
		_in  = _socket.getInputStream();
		_out = _socket.getOutputStream();
	}


	/**
	* Execute a DLP-command and return the result.
	*/	
	private synchronized byte[] dlp_exec(int cmd, int arg, final byte[] msg)
		throws DlpException
	{
		int i;

		byte[] exec_buf = new byte[DLP_BUF_SIZE];
	    exec_buf[0] = (byte)cmd;

	  	if ((msg != null) && (arg != 0) && (msg.length != 0))
	  	{
    		System.arraycopy(msg, 0, exec_buf, 6, msg.length);
    		exec_buf[1] = (byte)1;
    		exec_buf[2] = (byte)(arg | 0x80);
	    	exec_buf[3] = (byte)0;
    		PackUtils.setShort(exec_buf, 4, msg.length);
    		i = msg.length + 6;
		} 
		else 
		{
    		exec_buf[1] = (byte)0;
    		i = 2;
	  	}

		try
		{
			_out.write(exec_buf, 0, i);
		  	i = _in.read(exec_buf, 0, DLP_BUF_SIZE);
		} 
		catch( java.io.IOException ioe )
		{
      		throw new DlpException(1);
		}

	  	int err = PackUtils.getShort(exec_buf, 2);

	  	if (err != 0) 
	  	{
			if (err == 18)
				throw new com.syncbuilder.sync.CancelSyncException();
			else
	    		throw new DlpException(err);
	  	}

	  	if (exec_buf[0] != (byte)(cmd | 0x80)) 
		{ /* received wrong response */
      		throw new DlpException(1);
	  	}

	  	if (exec_buf[1] == (byte)0) /* no return blocks */
    		return null; 

		byte[] result = null;

		/* assume only one return block */
		if ((exec_buf[4] & 0xC0) == 0xC0) 
		{ /* Long arg */
	  		i = PackUtils.getLong(exec_buf, 6);
			result = new byte[i];			
  			System.arraycopy(exec_buf, 10, result, 0, i);
	  	} 
		else if ((exec_buf[4] & 0x80) != 0) 
    	{ /* Short arg */
	  		i = PackUtils.getShort(exec_buf, 6);
			result = new byte[i];			
  			System.arraycopy(exec_buf, 8, result, 0, i);
	    }
	    else 
	    { /* Tiny arg */
  			i = exec_buf[5] & 0xff;
			result = new byte[i];			
  			System.arraycopy(exec_buf, 6, result, 0, i);
	    }

    	return result;
	}


	/**
	* This conversion function is strictly for use within the DLP layer.
	* This particular date/time format does not occur anywhere else within the
	* Palm OS or its communications. 
	*/
	private static java.util.Date getDate(final byte[] data, int offset)
		throws ArrayIndexOutOfBoundsException
	{
    	int sec   = data[offset + 6] & 0xff;
    	int min   = data[offset + 5] & 0xff;
    	int hour  = data[offset + 4] & 0xff;
    	int mday  = data[offset + 3] & 0xff;
    	int mon   = (data[offset + 2] & 0xff) - 1;
    	int year  = (((data[offset + 0] & 0xff)<<8) | (data[offset + 1] & 0xff)) - 1900;

    	return new java.util.Date(year, mon, mday, hour, min, sec);
	}

	/**
	* Error checking. Throws a DlpException when the length of the array is shorter than the
	* specified minimum length.
	*/
	private static void _expect(byte[] b, int minlen)
		throws DlpException
	{
		if (b == null)
		{
			if (minlen == 0)
				return;
			else
				throw new DlpException(1);
			
		}
		if (b.length < minlen)
			throw new DlpException(1);
	}

	private final java.io.InputStream  _in;
	private final java.io.OutputStream _out;
	private final com.syncbuilder.sync.Socket _socket;

	private final static int DLP_BUF_SIZE = 0xffff; // Is the 64KB limit still there in the PalmIII???
}
