/* com/syncbuilder/service/ServiceBase.java:
*
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: ServiceBase.java,v 1.5 1998/10/25 14:09:31 christ Exp $ */

package com.syncbuilder.service;

/**
* This class is meant to be a suitable basis for most implementations
* of a service that implements ServiceControlListener.<br>
* You will only have to implement the service()-method, which will be
* invoked continuously in the service's main loop.<br>
* You may also implement the init()-method which will be invoked anytime
* the service is started, and the destroy()-method which will be invoked
* anytime the service is stopped.
* <p>
* See the Java Tutorial from JavaSoft if you wish to understand the
* way in which I start, stop, suspend, and resume the Thread.
* <p>
* Another very valuable facility of ServiceBase is its time-slicing facility.
* Any thread you register with the ServiceBase through the addThread()-method
* will be added to a pool of threads which are controlled through a time-slicing
* scheduler. This will avoid thread-starvation on systems which use green-threads.
*
* @author Tilo Christ
*/
public abstract class ServiceBase
    implements ServiceControlListener, Runnable
{
	public ServiceBase()
	{
		_queue_adaptor = new ServiceStateQueueAdaptor();
	}

	/**
	* This is the implementation of the run()-method from the
	* Runnable-interface.<br>
	* It will go through three distinct stages:<br>
	* 1. Initialization (invocation of init())<br>
	* 2. Providing service, until stopped (continuous invocations of service())<br>
	* 3. Release of occupied resources (invocation of destroy())<br>
	*/
	public void run()
	{
		System.err.println("ServiceBase.run()");

		try
		{
			init();
		}
		catch(Exception e)
		{
			// Initialization failed!!! What do we do now???
		}

		/* Service started */
		_notifyServiceStateListeners(ServiceStateEvent.RUNNING, "Service started");

		Thread this_thread = Thread.currentThread();

		// ------------------------- Loop here, while we are not stopped ---------------------------------
		while (this_thread == _service_thread)
		{
		
			// ------------------------- Loop here, if we are suspended ---------------------------------- 
			synchronized(this)
			{
				boolean is_suspended = false;
			
				if (_thread_suspended)
				{
					is_suspended = true;
					
					/* Service suspended */
					_notifyServiceStateListeners(ServiceStateEvent.SUSPENDED, "Service suspended");
				}
				
				while(_thread_suspended)
				{
					try
					{ 
						wait(); // We will be notified, if the thread shall no longer be suspended
					}
					catch(InterruptedException ie)
					{ // Perfectly normal, no special action req'd 
					}
				}

				if (is_suspended)
				{
					/* Service resumed */
					_notifyServiceStateListeners(ServiceStateEvent.RUNNING, "Service resumed");
				}
			}
			// ------------------------- Loop here, if we are suspended (End) ---------------------------- 

			if (this_thread != _service_thread)   // Accelerate exit from the loop, if we have been stopped
				break;
			
			// Perform the service (who really cares about all that other stuff!?!?)
			try
			{
				service();
			}
			catch(Exception e)
			{
				// Service execution failed
			}
		}
		// ------------------------- Loop here, while we are not stopped (End) ---------------------------

		try
		{
			destroy();
		}
		catch(Exception e)
		{
			// Release of resources failed
		}
		
		/* Service stopped */
		_notifyServiceStateListeners(ServiceStateEvent.STOPPED, "Service stopped");
	}


	/**
	* Temporarily halt the service. This will not release 
	* any resources and the service will be able to resume
	* operation very quickly.
	*/
	public synchronized void suspend()
	{
		if (_service_thread == null)
			return;
			
		_thread_suspended = true;
	}

	/**
	* Resume operation of the service after it has been
	* suspended. This is a very quick operation, because
	* all the resources that the service needs are still
	* allocated.
	*/
	public synchronized void resume()
	{
		if (_service_thread == null)
			return;
	
		if (_thread_suspended)
		{
			_thread_suspended = false;
			notify();
		}
	}

	/**
	* Start the service. This will switch the service into
	* a fully operational state. 
	* <br>
	* Starting a service will reserve all resources for that
	* service and might therefore be an expensive operation.
	*/
	public synchronized void start()
	{
		System.err.println("ServiceBase.start()");
	
		if (_service_thread == null)
		{
			_service_thread = new Thread(this);
			System.err.println("_service_thread.start()");
			_service_thread.start();

			addThread(_service_thread);
		}

		if (!_time_slicer_running)
		{
			_time_slicer_running = true;
			_time_slicer.start();
		}
	}

	/**
	* Stop the service. This will switch the service into
	* a completely unready state. It will release all allocated
	* resources and it will have to be restarted in order to
	* run again. Restarting the service from a stopped state
	* may be an expensive operation.
	*/
	public synchronized void stop()
	{
		_service_thread = null;

		if (_time_slicer_running)
		{
			_time_slicer_running = false;
			_time_slicer.stop();
		}
	
		if (_thread_suspended)
		{
			_thread_suspended = false;
			notify();
		}
	}

	/**
	* Add a thread to the time-slicing scheduler. The time-slicer
	* will take complete control over the threads priority, and 
	* you should NOT touch the priority after adding the thread
	* to the time-slicer.
	*/
	public void addThread(Thread thread)
	{
		_time_slicer.addThread(thread);
	}


	public synchronized void addServiceStateListener(ServiceStateListener sscl)
	{
		_queue_adaptor.addServiceStateListener(sscl);
	}

	public void removeServiceStateListener(ServiceStateListener sscl)
	{
		_queue_adaptor.removeServiceStateListener(sscl);
	}


	/**
	* This method is invoked anytime the service is started.
	* It should allocate all the resources which the service
	* needs.
	*/
	protected void init()
		throws Exception
	{
	}
	
	/**
	* This method is invoked anytime the service is stopped.
	* It should free all the resources which the service
	* held.
	*/
	protected void destroy()
		throws Exception
	{
	}

	/**
	* This method is invoked periodically while the service is running.
	* You need to implement the services functionality here.<br>
	* The method should be blocking for a limited amount of time (e.g. 500ms) 
	* and then return to the caller.<br>
	* If you are blocking for too short the service will hog CPU cycles,
	* if you are blocking for too long the service will become unresponsive
	* to control-events.
	*/
	protected abstract void service()
		throws Exception;


	protected void finalize()
		throws Throwable
	{
		_queue_adaptor.destroy();
		_queue_adaptor = null;
	}


	private void _notifyServiceStateListeners(int state, String message)
	{
		_notifyServiceStateListeners(state, message, ServiceMessage.LVL_NORMAL);
	}

	private void _notifyServiceStateListeners(int state, String message, int level)
	{
		_queue_adaptor.stateChanged(new ServiceStateEvent(this, null, state, new ServiceMessage(message, level)));
	}

	private Thread _service_thread = null;
	private boolean _thread_suspended = false;  // ONLY true if thread is started and suspended.

	private com.syncbuilder.util.TimeSlicer _time_slicer = new com.syncbuilder.util.TimeSlicer(Thread.NORM_PRIORITY);
	private boolean _time_slicer_running = false;

	private ServiceStateQueueAdaptor _queue_adaptor = null;
}
