/* com/syncbuilder/storage/CategoryAppBlock.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: CategoryAppBlock.java,v 1.1 1998/09/16 08:57:36 christ Exp $ */

package com.syncbuilder.storage;

import java.io.*;


/**
* The foundation of all AppBlocks which can store categories.
* The PalmOS has the built-in capability to assign each record
* of a database to one of 16 categories. The names of these
* categories are stored in an AppBlock which is associated with that database. 
* The name of each category is limited to a length of 15 characters.
*
* @author Kenneth Albanowski, changes by Tilo Christ
*/
public class CategoryAppBlock 
	extends AppBlock 
{
	public CategoryAppBlock() 
	{
		super();
	}

	public CategoryAppBlock(byte[] contents) 
	{
		super(contents);
	}
		
	public void unpack(byte[] data)
		throws java.lang.IllegalArgumentException
	{
		if (data.length < LENGTH)
			throw new IllegalArgumentException("data-array too short!");

		this.fill();  // prepare the arrays

		int r = ((data[4] & 0xff) << 8) + (data[5] & 0xff);
		
		for (int i=0;i<16;i++) 
		{
		    if ((r & (1<<i)) != 0)
		    	categoryRenamed[i] = true;
    		else
      			categoryRenamed[i] = false;
  		}

		for (int i=0;i<16;i++)
		{
			categoryName[i] = com.syncbuilder.util.PackUtils.getZeroTermString(data, 2+i*16, null);
		}
		
		for (int i=0;i<16;i++)
		{
			categoryID[i] = data[2 + 16*16 + i] & 0xff;
		}
		
		categoryLastUniqueID = data[2+ 16*16 + 16] & 0xff;
	}
	
	public byte[] pack()
	{
		byte[] buf = new byte[LENGTH];

		packCategories(buf);
		
		return buf;
    }


	public void fill() 
	{
		categoryRenamed = new boolean[16];
		categoryLastUniqueID = 0;

		categoryName = new String[16];
		for (int i = 0; i < 16; i++)
			categoryName[i] = new String();

		categoryID = new int[16];
	}

	
	/**
	* Test for equality with another Object.
	* Mostly used for testing.
	*/
	public boolean equals(Object other)
	{
		if (!(other instanceof com.syncbuilder.storage.CategoryAppBlock))
			return false;

		CategoryAppBlock other2 = (CategoryAppBlock)other;
			
		if (categoryLastUniqueID != other2.categoryLastUniqueID)
			return false;
			
		for (int i=0;i<16;i++)
		{
			if (categoryRenamed[i] != other2.categoryRenamed[i])
				return false;
			if (categoryID[i] != other2.categoryID[i])
				return false;
			if (!categoryName[i].equals(other2.categoryName[i]))
				return false;
		}

		return true;
	}

		
	public String describe() 
	{
		StringBuffer c;

		if ((categoryName == null) || (categoryID == null) || (categoryRenamed == null)) 
		{
			c = new StringBuffer("no categories");
		} 
		else 
		{
			c = new StringBuffer("name             id\trenamed\n");
			for (int i=0;i<categoryName.length;i++) 
			{
				if (categoryName[i].length() > 0) 
				{
					c.append(categoryName[i]);
					for(int j=16;j>categoryName[i].length();j--)
						c.append(" ");
					c.append(" ");
					c.append(categoryID[i]);
					c.append("\t");
					c.append(categoryRenamed[i]);
					c.append("\n");
				}
			}
		}
		return "categoryLastUniqueID="+categoryLastUniqueID+
				"\n" + c;
	}

	/** The fixed-length size of the native representation of the CategoryAppBlock */
	public final static int LENGTH = 2 + 16*16 + 16 + 4;

	public boolean[] categoryRenamed;
	public int categoryLastUniqueID;
	public String[] categoryName;
	public int[] categoryID;


	protected void packCategories(byte[] buf)
	{
		int r = 0;

	    for (int i=0;i<16;i++) 
		{
    		if (categoryRenamed[i])
      			r |= (1<<i);
		}
		
		buf[0] = (byte)((r >> 8) & 0xff);
		buf[1] = (byte)(r & 0xff);
		
		for (int i=0;i<16;i++)
		{
			com.syncbuilder.util.PackUtils.setZeroTermString(buf, 2 + i*16, categoryName[i]);
		}

		for (int i=0;i<16;i++)
		{
			buf[2 + 16*16 + i] = (byte)categoryID[i];
		}

		buf[2 + 16*16 + 16] = (byte)categoryLastUniqueID;
		
		buf[2 + 16*16 + 16 + 1] = (byte)0;  // gap-fill
		buf[2 + 16*16 + 16 + 2] = (byte)0;  // gap-fill
		buf[2 + 16*16 + 16 + 3] = (byte)0;  // gap-fill
	}
}
