/* com/syncbuilder/storage/Char4.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Char4.java,v 1.1 1998/09/16 08:57:36 christ Exp $ */

package com.syncbuilder.storage;


/**
* The representation of a 4-character identifier. 
* These identifiers are needed for the Creator-IDs and the database types.
* They need special treatment, because they are 4-byte integers, where each
* byte is an 8-bit character for the sake of readability.<p>
* Instances of this class are immutable.
*
* @author Kenneth Albanowski, changes by Tilo Christ
*/
final public class Char4 
{
	// Do NOT touch this one! It is req'd by the native code interface!
	/**
	* Create a Char4 from a 32-bit integer.
	* 
	* @param id the 32-bit integer
	*/
	public Char4(int id) 
	{
		value = id;
		b = new byte[4];
		b[3] = (byte)(id & 0xff); id >>= 8;
		b[2] = (byte)(id & 0xff); id >>= 8;
		b[1] = (byte)(id & 0xff); id >>= 8;
		b[0] = (byte)(id & 0xff); id >>= 8;
	}

	/**
	* Create a Char4 from a String of 4 characters.
	* 
	* @param id the String
	* @throws java.lang.IllegalArgumentException The String is not 4 characters long
	*/
	public Char4(String id) 
		throws java.lang.IllegalArgumentException
	{
		if (id.length() != 4)
		{
			throw new java.lang.IllegalArgumentException("Length <> 4 bytes!");
		}  

		byte[] by = id.getBytes();

		this.b = (byte[])by.clone();

		value = (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];

	}

	/**
	* Create a Char4 from an array of 4 bytes.
	* 
	* @param id the array
	* @throws java.lang.IllegalArgumentException The array is not 4 bytes long
	*/
	public Char4(byte[] b) 
		throws java.lang.IllegalArgumentException
	{
		if (b.length != 4)
		{
			throw new java.lang.IllegalArgumentException("Length <> 4 bytes!");
		}  

		this.b = (byte[])b.clone();

		value = (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];
	}

	/**
	* Get the 32-bit integer-value of the Char4.
	*/
	public int getInt() 
	{
		return value;
	}
	
	/**
	* Get the array of 4 bytes from the Char4.
	*/
	public byte[] getBytes() 
	{
		return b;
	}

	/**
	* Get the 4-character String of the Char4.
	*/
	public String getString() 
	{
		return new String(b, 0, b.length);
	}
	
	/**
	* Is the Char4 equal to another one?
	*/
	public boolean equals(Char4 other) 
	{
		return ( value == other.getInt() );
	}
	
	/**
	* Return the Char4's value as a pretty human-readable string.
	*/
	public String toString() {
		return com.syncbuilder.util.Util.prettyPrint(b);
	}

	
	private int value;
	private byte[] b;
}
