/* com/syncbuilder/storage/Database.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Database.java,v 1.3 1998/12/09 09:56:17 christ Exp $ */

package com.syncbuilder.storage;

import java.io.*;

import com.syncbuilder.sync.DlpException;
import com.syncbuilder.sync.Dlp;
import com.syncbuilder.sync.Link;
import com.syncbuilder.internal.core.constants;

/** 
 * A representation of an open database.
 * Database is assisted by DatabaseImpl, which provides the neccessary knowledge to
 * treat databases from different applications properly. This is a delegation-/bridge pattern.<p>
 * Database-objects are supposed to be obtained through the methods of the Dlp
 * class (such as openDB(), openStockDB(), createDB()).
 *
 * @see com.syncbuilder.sync.Dlp
 * @see com.syncbuilder.storage.DatabaseImpl
 * @author Kenneth Albanowski
 */
public class Database 
{
//=============================== AppBlock ====================================

	public AppBlock getAppBlock() 
		throws DlpException 
	{
		return _dlp.dlp_ReadAppBlock(_handle, _dbimpl);
	}

	public void setAppBlock(AppBlock appblock) 
		throws DlpException 
	{
		_dlp.dlp_WriteAppBlock(_handle, appblock);
	}

	public AppBlock createAppBlock() 
	{
		return _dbimpl.createAppBlock();
	}

//=============================== SortBlock ====================================

	public SortBlock getSortBlock() throws DlpException {
		return _dlp.dlp_ReadSortBlock(_handle, _dbimpl);
	}

	public void setSortBlock(SortBlock sortblock) 
		throws DlpException 
	{
		_dlp.dlp_WriteSortBlock(_handle, sortblock);
	}

	public SortBlock createSortBlock() {
		return _dbimpl.createSortBlock();
	}


//=============================== Record ====================================

	public Record getRecord(int index) 
		throws DlpException 
	{
		return _dlp.dlp_ReadRecordByIndex(_handle, index, _dbimpl);
	}

	public Record getRecord(RecordID id) 
		throws DlpException 
	{
		return _dlp.dlp_ReadRecordByID(_handle, id, _dbimpl);
	}

	public Record getNextRecord(int category) 
		throws DlpException 
	{
		return _dlp.dlp_ReadNextRecInCategory(_handle, category, _dbimpl);
	}

	public Record getNextModRecord() 
		throws DlpException 
	{
		return _dlp.dlp_ReadNextModifiedRec(_handle, _dbimpl);
	}

	public Record getNextModRecord(int category) 
		throws DlpException 
	{
		return _dlp.dlp_ReadNextModifiedRecInCategory(_handle, category, _dbimpl);
	}


	/**
	* Writes a record to the database.
	* If the Record had an unspecified RecordID (0) before
	* it will be assigned a valid RecordID.
	*/
	public void putRecord(Record record) 
		throws DlpException 
	{
		_dlp.dlp_WriteRecord(_handle, record);
	}


	/**
	* Create a new Record with an unspecified RecordID (0).
	*/
	public Record createRecord() 
	{
		return _dbimpl.createRecord();
	}

	/**
	* Create a new Record with the specified RecordID.
	*/
	public Record createRecord(RecordID id) 
	{
		return _dbimpl.createRecord(id);
	}

	/**
	* Delete the record with the specified UID from the database.
	* 
	* @param uid the UID (not the index!) of the record
	*/
	public void deleteRecord(RecordID uid) 
		throws DlpException 
	{
		_dlp.dlp_DeleteRecord(_handle, false, uid);
	}


	/**
	* Delete ALL records in the database.
	* Use with caution!
	*/
	public void deleteRecords() 
		throws DlpException 
	{
		_dlp.dlp_DeleteRecord(_handle, true, null);
	}


	public RecordID[] getRecordIDs()	
		throws DlpException
	{ 
		return getRecordIDs(false);	
	}

	public RecordID[] getRecordIDs(boolean sort) 	
		throws DlpException
	{	
		return getRecordIDs(sort, 0); 	
	}

	public RecordID[] getRecordIDs(boolean sort, int start)		
		throws DlpException
	{	
		return getRecordIDs(sort, start, 0xffff);	
	}

	public RecordID[] getRecordIDs(boolean sort, int start, int max)
		throws DlpException
	{
		return _dlp.dlp_ReadRecordIDList(_handle, sort, start, max);
	}


	/**
	* Deletes all records which are marked as archived or deleted
	*/
	public void purgeRecords() 
		throws DlpException 
	{
		_dlp.dlp_CleanUpDatabase(_handle);
	}



//=============================== Resource ====================================

	public Resource createResource() {
		return _dbimpl.createResource();
	}

	public Resource createResource(Char4 type, int id) {
		return _dbimpl.createResource(type, id);
	}

	public Resource getResource(int index) 
		throws DlpException 
	{
		return _dlp.dlp_ReadResourceByIndex(_handle, index, _dbimpl);
	}

	public Resource getResource(Char4 type, int id) 
		throws DlpException 
	{
		return _dlp.dlp_ReadResourceByType(_handle, type, id, _dbimpl);
	}

	public void putResource(Resource resource) 
		throws DlpException 
	{
		_dlp.dlp_WriteResource(_handle, resource);
	}

	public void deleteResource(Char4 type, int id) 
		throws DlpException 
	{
		_dlp.dlp_DeleteResource(_handle, false, type, id);
	}


	/**
	* Delete ALL resources.
	* Use with caution!
	*/
	public void deleteResources() 
		throws DlpException 
	{
		_dlp.dlp_DeleteResource(_handle, true, null, 0);
	}


//=============================== Category ====================================

	public void deleteCategory(int category) 
		throws DlpException 
	{
		_dlp.dlp_DeleteCategory(_handle, category);
	}

	public void moveCategory(int from, int to) 
		throws DlpException 
	{
		_dlp.dlp_MoveCategory(_handle, from, to);
	}

//=============================== Preference ====================================

	
	public Pref getPref(int id)
	  throws DlpException, IOException
	{	
		return getPref(id, true); 
	}


	/**
	* Get one of the preferences which are associated with the same app as the database.
	*/
	public Pref getPref(int id, boolean backup)
		throws DlpException, NoCreatorException, IOException
	{
		if (_creator == null)
			throw new NoCreatorException();

		if (_link.getProtocolVersion() < 0x101)
			_dlp.dlp_CloseDB(_handle);

		Pref result = _dlp.dlp_ReadAppPreference(_creator, id, backup, _dbimpl);

		if (_link.getProtocolVersion() < 0x101)
			_handle = _dlp.reopenDB( _dbname, _dbmode, _dbcard);

		return result;
	}


	public void setPref(Pref pref)
		throws DlpException, NoCreatorException, IOException
	{
		if (_creator == null)
			throw new NoCreatorException();

		if (_link.getProtocolVersion() < 0x101)
			_dlp.dlp_CloseDB(_handle);

		_dlp.dlp_WriteAppPreference(pref);

		if (_link.getProtocolVersion() < 0x101)
			_handle = _dlp.reopenDB(_dbname, _dbmode, _dbcard);
	}


	public Pref createPref(int id)
	{
		return _dbimpl.createPref(null, id, 1, true);
	}

	public Pref createPref(int id, int version, boolean backup)
	{
		return _dbimpl.createPref(null, id, version, backup);
	}


//=============================== Miscellaneous ====================================

	/**
	* Closes the database. 
	* The Database-object shall no longer be used after the invocation of this method!
	*/
	synchronized public void close() 
		throws DlpException 
	{ 
		/* this method should be idempotent */
		if (_handle != 0)
			_dlp.dlp_CloseDB(_handle);
			
		_handle = 0;
	}


	/**
	* For record databases, reset all dirty flags. For both record and
	* resource databases, set the last sync time to now. 
	*/
	public void resetFlags() 
		throws DlpException 
	{
		_dlp.dlp_ResetSyncFlags(_handle);
	}

	public void resetNext() 
		throws DlpException 
	{
		_dlp.dlp_ResetDBIndex(_handle);
	}

	/**
	* Return the number of records in the Database
	*/
	public int getRecordCount() 
		throws DlpException 
	{
		return _dlp.dlp_getRecordCount(_handle);
	}





	/* These constants are used to describe the mode in which
	  the database is to be opened */
	/** Opened for read access */
	public final static int OPEN_READ       = constants.dlpOpenRead;
	/** Opened for write access */
	public final static int OPEN_WRITE      = constants.dlpOpenWrite;
	/** Opened for exclusive access */
	public final static int OPEN_EXCLUSIVE  = constants.dlpOpenExclusive;
	
	/** Opened with default settings (read/write, exclusive) */
	public final static int OPEN_DEFAULT = OPEN_READ|OPEN_WRITE|OPEN_EXCLUSIVE;
	
	
	/**
	* Construct a Database object.<br>
	* <b>!!! This constructor shall only be used by Dlp!!!</b>
	* This constructor tries to figure out the proper version of DatabaseImpl
	* by querying the DatabaseImplFactory.
	*
	* @param link an instance of Link
	* @param handle a database-handle as returned by DlpImpl.openDB
	* @param creator the creator ID of the database
	* @param dbname the name of the database
	* @param dbmode the mode of the database, use the Database.OPEN_XXX-flags
	* @param dbcard the card on which the database is located.
	*/
	public Database(Link link, int handle, Char4 creator, String dbname, int dbmode, int dbcard) 
	{
		this._link = link;
		this._dlp = link.getDlp();
		this._handle = handle;
		this._creator = creator;
		this._dbname = dbname;
		this._dbmode = dbmode;
		this._dbcard = dbcard;

		_dbimpl = DatabaseImplFactory.findDatabaseImplByName(dbname);

		if (_dbimpl == null) 
		{
			_dbimpl = new com.syncbuilder.storage.generic.DatabaseImpl(creator, dbname);
		}
		else
		{
			_creator = _dbimpl.getCreatorID();
		}
	}


	protected void finalize() 
	  throws DlpException 
	{
		this.close();
	}


	private DatabaseImpl _dbimpl;

	private Link _link;
	private Dlp _dlp;
	private int _handle;
	private Char4 _creator;
	private String _dbname;
	private int _dbmode;
	private int _dbcard;
}

