/* com/syncbuilder/storage/Record.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Record.java,v 1.1 1998/09/16 08:57:38 christ Exp $ */

package com.syncbuilder.storage;

import com.syncbuilder.internal.core.constants;

/** 
 * The base-class for representations of a database-record.
 * This type is meant for storage-databases. 
 * The content of resource-databases is represented by com.syncbuilder.storage.Resource.
 * 
 * @see com.syncbuilder.storage.Resource
 * @author Kenneth Albanowski, changes by Tilo Christ
 */
abstract public class Record 
	extends Block 
{
	public Record() 
	{
		super();
	}

	/**
	* Create a Record with the specified content and attributes.
	* The content which you specify as raw bytes is automatically 
	* processed by the <code>unpack()</code>-method to distribute it to the 
	* properties of the Record.
	*
	* @param contents the content in the native format of the PalmPilot application
	* @param id the ID of the Record, or <code>null</code> if the ID shall be assigned by the PalmOS during HotSync.
	* @param index the index of the Record
	* @param attr the attributes. Use the DELETED, MODIFIED, BUSY, SECRET, and ARCHIVED flags.
	* @param category the category of the Record. Use values from 0-15.
	*/
	public Record(byte[] contents, RecordID id, int index, int attr, int category) 
		throws IllegalArgumentException
	{
		if (category < 0 || category > 15)
			throw new IllegalArgumentException("Invalid category number.");

		if (id == null)
			id = new RecordID();

		this.id = id;
		this.index = index;
		this.deleted = ((attr & DELETED)!=0) ? true : false;
		this.modified = ((attr & MODIFIED)!=0) ? true : false;
		this.busy = ((attr & BUSY)!=0) ? true : false;
		this.secret = ((attr & SECRET)!=0) ? true : false;
		this.archived = ((attr & ARCHIVED)!=0) ? true : false;
		this.category = category;

		if (!isDeleted())
			this.unpack(contents);
	}

	/**
	* Initialize the Record. 
	* All attributes are set to false. Index will be 0. Category is 0.
	* ID will be assigned by the PalmOS when the Record is written to the device.
	*/
	public void fill() 
	{
		this.id = new RecordID();
		this.index = 0;
		this.category = 0;
		
		this.deleted = false;
		this.modified = false;
		this.busy = false;
		this.secret = false;
		this.archived = false;
	}

	public String describe() 
	{
		return " id "+id+", index "+index+", deleted "+deleted+", modified "+modified+
		", busy "+busy+", secret "+secret+", archived "+archived+", category "+ category;
	}


	/**
	* Determine the equality of this Record and some other Object.
	*/
	public boolean equals(Object obj)
	{
		if (obj == null)
			return false;
		
		if (!(obj instanceof com.syncbuilder.storage.Record))
			return false;
			
		com.syncbuilder.storage.Record other = (com.syncbuilder.storage.Record)obj;
		
		if (index != other.index)
			return false;
			
		if (category != other.category)
			return false;
			
		if (!id.equals(other.id))
			return false;
			
		if (this.getFlags() != other.getFlags())
			return false;

		return true;
	}

	public boolean isDeleted()
	{
		return deleted;
	}

	public boolean isModified()
	{
		return modified;
	}

	public boolean isBusy()
	{
		return busy;
	}

	public boolean isSecret()
	{
		return secret;
	}

	public boolean isArchived()
	{
		return archived;
	}

	public void setDeleted(boolean flag)
	{
		deleted = flag;
	}

	public void setModified(boolean flag)
	{
		modified = flag;
	}

	public void setBusy(boolean flag)
	{
		busy = flag;
	}

	public void setSecret(boolean flag)
	{
		secret = flag;
	}

	public void setArchived(boolean flag)
	{
		archived = flag;
	}

	/**
	* Get a numerical representation of the flags
	*/
	public int getFlags()
	{
		int flags = 0;
		
		if (deleted)
			flags |= DELETED;
		if (modified)
			flags |= MODIFIED;
		if (busy)
			flags |= BUSY;
		if (secret)
			flags |= SECRET;
		if (archived)
			flags |= ARCHIVED;

		return flags;
	}

	public int getIndex()
	{
		return index;
	}

	public void setIndex(int index)
	{
		this.index = index;
	}
	
	public RecordID getID()
	{
		return id;
	}
	
	public void setID(RecordID id)
	{
		if (id == null)
			id = new RecordID();
			
		this.id = id;
	}

	public int getCategory()
	{
		return category;
	}
	
	public void setCategory(int category)
		throws IllegalArgumentException
	{
		if (category < 0 || category > 15)
			throw new IllegalArgumentException("Invalid category number.");

		this.category = category;
	}

	public final static int DELETED   = constants.dlpRecAttrDeleted;
	public final static int MODIFIED  = constants.dlpRecAttrDirty;
	public final static int BUSY      = constants.dlpRecAttrBusy;
	public final static int SECRET    = constants.dlpRecAttrSecret;
	public final static int ARCHIVED  = constants.dlpRecAttrArchived;

	protected int index;
	protected int category;
	/** The unique ID of the record. This is guaranteed to be not null. */
	protected RecordID id;
	protected boolean deleted;
	protected boolean modified;
	protected boolean busy;
	protected boolean secret;
	protected boolean archived;
}
