/* com/syncbuilder/sync/Dlp.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
* Copyright (C) 1998, Robert Crawford
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Dlp.java,v 1.2 1998/12/09 09:45:55 christ Exp $ */

package com.syncbuilder.sync;

import java.io.*;
import com.syncbuilder.storage.*;
import com.syncbuilder.device.*;
import com.syncbuilder.internal.core.calls;
import com.syncbuilder.internal.core.constants;

/**
 * A representation of a DLP (HotSync)-connection.<p>
 * DLP is layered on top of a reliable stream and can initiate many operations
 * on the Pilot, in the areas of Database operations and global information
 * (system time, user ID, preferences).<p>
 *
 * From <b>Developing Palm OS 2.0 Applications</b>:<br>
 * <BLOCKQUOTE>
 *   The Desktop Link Protocol (DLP) provides remote access to Palm OS data
 *   storage and other subsystems.<p>
 *   
 *   DLP facilitates efficient data synchronization between desktop (PC,
 *   Macintosh, etc.) and Palm OS applications, database backup, installation
 *   of code patches, extensions, applications, and other databases ...
 * </BLOCKQUOTE>
 *
 * @see com.syncbuilder.sync.Socket
 * @author Kenneth Albanowski, Tilo Christ, Robert Crawford
 */
final public class Dlp 
{
	/**
	* Open a database by its name.
	* Database will be opened R/W, exclusive. It will be looked for on card 0 (built-in).
	* <br><b>Caution: </b>It is impossible to keep more than one database open at the same time!
	* This is a limitation of DLP and not of this framework. Use <code>closeDB()</code> to close databases.
    *
	* @param name the name of the database
	*/
	public Database openDB(String name) 
		throws DlpException 
	{ 
		return this.openDB(name, Database.OPEN_DEFAULT); 
	}

	/**
	* Open a database by its name.
	* Database will be looked for on card 0 (built-in).
	* <br><b>Caution: </b>It is impossible to keep more than one database open at the same time!
	* This is a limitation of DLP and not of this framework. Use <code>closeDB()</code> to close databases.
    *
	* @param name the name of the database
	* @param mode the mode in which the database will be opened. Use the Database.OPEN_*-constants.
	*/
	public Database openDB(String name, int mode) 
		throws DlpException 
	{ 
		return this.openDB(name, mode, CardInfo.DEFAULTCARD); 
	}

	/**
	* Open a database by its name. 
	* <br><b>Caution: </b>It is impossible to keep more than one database open at the same time!
	* This is a limitation of DLP and not of this framework. Use <code>closeDB()</code> to close databases.
    *
	* @param name the name of the database
	* @param mode the mode in which the database will be opened. Use the Database.OPEN_*-constants.
	* @param card the card where the database is located. The built-in card is CardInfo.DEFAULTCARD.
	*/
	public Database openDB(String name, int mode, int card) 
		throws DlpException 
	{
		Char4 creator = null;

		DatabaseImpl dbimpl = DatabaseImplFactory.findDatabaseImplByName(name);
		if (dbimpl != null)
		{
			creator = dbimpl.getCreatorID();
		}
		
		return openDB(creator, name, mode, card);
	}

	/**
	* Open a database by its name, and with information about its creator.
	* The creator ID will not be used to find the database, but to add more information
	* to the resulting Database object.
	* <br><b>Caution: </b>It is impossible to keep more than one database open at the same time!
	* This is a limitation of DLP and not of this framework. Use <code>closeDB()</code> to close databases.
    *
	* @param creator the creator ID of the database
	* @param name the name of the database
	* @param mode the mode in which the database will be opened. Use the Database.OPEN_*-constants.
	* @param card the card where the database is located. The built-in card is CardInfo.DEFAULTCARD.
	*/
	public Database openDB(Char4 creator, String name, int mode, int card)
		throws DlpException
	{
		int handle = _impl.openDB(name, mode, card);

		return new Database(_link, handle, creator, name, mode, card);
	}

	/**
	* Open a database with a known DatabaseImpl. 
	* <br><b>Caution: </b>It is impossible to keep more than one database open at the same time!
	* This is a limitation of DLP and not of this framework. Use <code>closeDB()</code> to close databases.
	*
	* @see com.syncbuilder.storage.memo.DatabaseImpl
	* @see com.syncbuilder.storage.mail.DatabaseImpl
	* @see com.syncbuilder.storage.todo.DatabaseImpl
	* @see com.syncbuilder.storage.address.DatabaseImpl
	* @see com.syncbuilder.storage.appointment.DatabaseImpl
	*/
	public Database openStockDB(DatabaseImpl dbimpl)
		throws DlpException
	{
		return this.openDB(dbimpl.getCreatorID(), dbimpl.getDBName(), Database.OPEN_DEFAULT, CardInfo.DEFAULTCARD);
	}


	/**
	* Finish the HotSync-process successfully. 
	* Invoking this method ends the life-cycle of the Dlp. It cannot be used afterwards.
	*/
    public void endHotSync() 
		throws DlpException, IOException 
	{
		endHotSync(END_NORMAL);
	}

	/**
	* Finish the HotSync-process with a specified status-code
	* Invoking this method ends the life-cycle of the Dlp. It cannot be used afterwards.
	*
	* @param status the way in which the HotSync ended (use Dlp.END_*-constants).
	*/
	public synchronized void endHotSync(int status) 
	  throws DlpException, IOException 
	{
		_impl.endHotSync(status);
		_socket.close();

		// These resources have become useless. Let's discard them!
		_socket = null;
		_impl = null;
	}

	public Database createDB(String name, Char4 creator, Char4 type, int flags, int version) 
		throws DlpException
    { 
		return createDB(name, creator, type, flags, version, CardInfo.DEFAULTCARD); 
	}

	public Database createDB(String name, Char4 creator, Char4 type, int flags, int version, int card) 
		throws DlpException
	{
		int handle = _impl.createDB(name, creator, type, card, flags, version);
		return new Database(_link, handle, creator, name, Database.OPEN_DEFAULT, card);
	}

	public void resetPilot() 
	  throws DlpException 
	{
		dlp_ResetSystem();
	}


	public CardInfo getCardInfo(int card)
		throws DlpException
	{
		return this.dlp_ReadStorageInfo(card);
	}

	public java.util.Date getTime()
		throws DlpException
	{
		return dlp_GetSysDateTime();
	}

	public UserInfo getUserInfo()
		throws DlpException
	{
		return _impl.readUserInfo();
	}

	public SysInfo getSysInfo()
		throws DlpException
	{
		return dlp_ReadSysInfo();
	}

	/**
	* Make the Palm device's display message switch from 
	* "Identifying User" to "Synchronizing"
	*/
	public void switchMessage()
		throws DlpException
	{
		_impl.switchMessage();
	}

	public DBInfo getDBInfo(int start, boolean RAM, boolean ROM, int card)
		throws DlpException
	{
		return _impl.getDBInfo(start, card, 
							  	(RAM ? constants.dlpDBListRAM : 0) |
							   	(ROM ? constants.dlpDBListROM : 0)
							   );
	}

	public int getFeature(Char4 creator, int id)
		throws DlpException
	{
		return dlp_ReadFeature(creator, id);
	}

	/**
	* Get a Pref object that is associated with a particular application.
	* Do not invoke this method while there is a database from the same
	* application open! This will lead to problems with Palm devices running
	* Palm OS 1.0! Use the getPref()-method from the Database class instead.
	*/
	public Pref getPref(Char4 creator, int id)
		throws DlpException
	{	
		return getPref(creator, id, true); 
	}

	/**
	* Get a Pref object that is associated with a particular application.
	* Do not invoke this method while there is a database from the same
	* application open! This will lead to problems with Palm devices running
	* Palm OS 1.0! Use the getPref()-method from the Database class instead.
	*/
	public Pref getPref(Char4 creator, int id, boolean backup)
		throws DlpException, NoCreatorException
	{
		if (creator == null)
			throw new NoCreatorException();

		DatabaseImpl dbimpl = DatabaseImplFactory.findDatabaseImplByCreator(creator);

		if (dbimpl == null)
			dbimpl = new com.syncbuilder.storage.generic.DatabaseImpl(creator, null);

		return dlp_ReadAppPreference(creator, id, backup, dbimpl); 
	}

	/**
	* Set a Pref object that is associated with a particular application.
	* Do not invoke this method while there is a database from the same
	* application open! This will lead to problems with Palm devices running
	* Palm OS 1.0! Use the setPref()-method from the Database class instead.
	*/
	public void setPref(Pref pref)
		throws DlpException
	{
		dlp_WriteAppPreference(pref);
	}

    /**
     * A factory method for a <code>Pref</code> object.
	 * The new Pref will be a specialization of com.syncbuilder.storage.Pref that matches
	 * the specified creator. The Pref's version will be set to 1.
	 * It's backup-flag will be set.
	 *
	 * @param creator the creator ID 
	 * @param 
	 * @return	a <code>Pref</code> object.
     */
	public Pref createPref(Char4 creator, int id)
		throws DlpException, NoCreatorException
	{	
		return createPref(creator, id, 1); 
	}
	
    /**
     * A factory method for a <code>Pref</code> object.
	 * The new Pref will be a specialization of com.syncbuilder.storage.Pref that matches
	 * the specified creator and id. It's backup-flag will be set.
	 *
	 * @return	a <code>Pref</code> object.
     */
	public Pref createPref(Char4 creator, int id, int version)
		throws DlpException, NoCreatorException
	{	
		return createPref(creator, id, version, true); 
	}
	
	public Pref createPref(Char4 creator, int id, int version, boolean backup)
		throws DlpException, NoCreatorException
	{
		if (creator == null)
			throw new NoCreatorException();

		DatabaseImpl dbimpl = DatabaseImplFactory.findDatabaseImplByCreator(creator);

		if (dbimpl == null)
			dbimpl = new com.syncbuilder.storage.generic.DatabaseImpl(creator, null);

		return dbimpl.createPref(null, id, version, backup); 
	}

	public NetSyncInfo getNetSyncInfo()
		throws DlpException
	{
		return dlp_ReadNetSyncInfo();
	}

	public void setNetSyncInfo(NetSyncInfo info)
		throws DlpException
	{
		dlp_WriteNetSyncInfo(info);
	}

	public void addSyncLogEntry(String message) 
		throws DlpException 
	{
		_impl.addSyncLogEntry(message);
	}

	public byte[] callApplication(Char4 creator, int type, int action, byte[] data, int[] retcode)
		throws DlpException
	{
		return dlp_CallApplication(creator, type, action, data, retcode);
	}

	public void deleteDB(String dbname) 
		throws DlpException
	{	
		deleteDB(dbname, CardInfo.DEFAULTCARD); 
	}
	
	public void deleteDB(String dbname, int card) 
		throws DlpException 
	{
		dlp_DeleteDB(card, dbname);
	}


	/*
 	* The following constants can be passed to the endHotSync-method in order to specify 
	* the way the HotSync ended.
    */

	/** HotSync ended normally */
	public final static int END_NORMAL         = constants.dlpEndCodeNormal;
	/** HotSync ended because of low memory on the Pilot */
	public final static int END_OUT_OF_MEMORY  = constants.dlpEndCodeOutOfMemory;
	/** HotSync ended because user cancelled the operation */
	public final static int END_USER_CANCELLED = constants.dlpEndCodeUserCan;
	/** HotSync ended with an unspecified problem */
	public final static int END_OTHER          = constants.dlpEndCodeOther;
	


	/******************************************************************************************
	* The following methods are very low-level. 
	* You should use the high-level versions where possible. 
	******************************************************************************************/

//=======================  Database ===========================================================

	public synchronized void dlp_CloseDB(int handle) throws DlpException
		{ _impl.closeDB(handle); }

	/** Open a database by its name. Do not construct a Database object, but only return the handle. */
	public synchronized int reopenDB(String name, int mode, int card) throws DlpException
		{ return _impl.openDB(name, mode, card); }

	public synchronized int dlp_getRecordCount(int handle) throws DlpException
		{ return _impl.getRecordCount( handle ); }

	synchronized DBInfo dlp_ReadDBList(int card, int flags, int start) throws DlpException
		{ return _impl.getDBInfo(start, card, flags); }

	synchronized int dlp_DeleteDB(int card, String name) throws DlpException
		{ return calls.dlp_DeleteDB(_socket.getDescriptor(), card, name); }

	public synchronized int dlp_CleanUpDatabase(int handle) throws DlpException
		{ return calls.dlp_CleanUpDatabase(_socket.getDescriptor(), handle); }

	public synchronized int dlp_MoveCategory(int handle, int from, int to) throws DlpException
		{ return calls.dlp_MoveCategory(_socket.getDescriptor(), handle, from, to); }

	public synchronized int dlp_DeleteCategory(int handle, int category) throws DlpException
		{ return calls.dlp_DeleteCategory(_socket.getDescriptor(), handle, category); }

	public synchronized int dlp_ResetDBIndex(int handle) throws DlpException
		{ return calls.dlp_ResetDBIndex(_socket.getDescriptor(), handle); }


//=======================  Record =============================================================

	public synchronized Record dlp_ReadRecordByIndex(int handle, int index, DatabaseImpl dbclass) throws DlpException
		{ return _impl.readRecordByIndex(handle, index, dbclass); }

	public synchronized Record dlp_ReadRecordByID(int handle, RecordID id, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadRecordByID(_socket.getDescriptor(), handle, id, dbClass); }

	public synchronized void dlp_WriteRecord(int handle, Record record) throws DlpException
	{ _impl.writeRecord(handle, record); }

	public synchronized Record dlp_ReadNextRecInCategory(int handle, int category, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadNextRecInCategory(_socket.getDescriptor(), handle, category, dbClass); }

	public synchronized RecordID[] dlp_ReadRecordIDList(int handle, boolean sort, int start, int max) throws DlpException
		{ return calls.dlp_ReadRecordIDList(_socket.getDescriptor(), handle, sort, start, max); }

	public synchronized int dlp_DeleteRecord(int handle, boolean all, RecordID id) throws DlpException
		{ _impl.deleteRecord(handle, all, id); return 0; }

	public synchronized Record dlp_ReadNextModifiedRec(int handle, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadNextModifiedRec(_socket.getDescriptor(), handle, dbClass); }

	public synchronized Record dlp_ReadNextModifiedRecInCategory(int handle, int category, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadNextModifiedRecInCategory(_socket.getDescriptor(), handle, category, dbClass); }

//=======================  Resource =============================================================

	public synchronized int dlp_WriteResource(int handle, Resource resource) throws DlpException
		{ return calls.dlp_WriteResource(_socket.getDescriptor(), handle, resource); }

	public synchronized Resource dlp_ReadResourceByType(int handle, Char4 type, int id, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadResourceByType(_socket.getDescriptor(), handle, type, id, dbClass); }

	public synchronized Resource dlp_ReadResourceByIndex(int handle, int index, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadResourceByIndex(_socket.getDescriptor(), handle, index, dbClass); }

	public synchronized int dlp_DeleteResource(int handle, boolean all, Char4 type, int id) throws DlpException
		{ return calls.dlp_DeleteResource(_socket.getDescriptor(), handle, all, type, id); }

//=======================  Preference & Feature ==================================================

	public synchronized Pref dlp_ReadAppPreference(Char4 creator, int id, boolean backup, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadAppPreference(_socket.getDescriptor(), creator, id, backup, dbClass); }

	public synchronized int dlp_WriteAppPreference(Pref pref) throws DlpException
		{ return calls.dlp_WriteAppPreference(_socket.getDescriptor(), pref); }

	/** This method is implemented in a nasty, undocumented way. Use it at your own discretion */
	synchronized int dlp_ReadFeature(Char4 creator, int id) throws DlpException
		{ return calls.dlp_ReadFeature(_socket.getDescriptor(), creator, id); }

//=======================  AppBlock & SortBlock ==================================================

	public synchronized AppBlock dlp_ReadAppBlock(int handle, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadAppBlock(_socket.getDescriptor(), handle, dbClass); }

	public synchronized SortBlock dlp_ReadSortBlock(int handle, DatabaseImpl dbClass) throws DlpException
		{ return calls.dlp_ReadSortBlock(_socket.getDescriptor(), handle, dbClass); }

	public synchronized int dlp_WriteAppBlock(int handle, AppBlock appblock) throws DlpException
		{ return calls.dlp_WriteAppBlock(_socket.getDescriptor(), handle, appblock); }

	public synchronized int dlp_WriteSortBlock(int handle, SortBlock sortblock) throws DlpException
		{ return calls.dlp_WriteSortBlock(_socket.getDescriptor(), handle, sortblock); }

//=======================  Palm device =========================================================

	synchronized CardInfo dlp_ReadStorageInfo(int card) throws DlpException
		{ return calls.dlp_ReadStorageInfo(_socket.getDescriptor(), card); }

	synchronized java.util.Date dlp_GetSysDateTime() throws DlpException
		{ return calls.dlp_GetSysDateTime(_socket.getDescriptor()); }

	synchronized int dlp_WriteUserInfo(UserInfo info) throws DlpException
		{ return calls.dlp_WriteUserInfo(_socket.getDescriptor(), info); }

	synchronized SysInfo dlp_ReadSysInfo() throws DlpException
		{ return calls.dlp_ReadSysInfo(_socket.getDescriptor()); }

	synchronized NetSyncInfo dlp_ReadNetSyncInfo() throws DlpException
		{ return calls.dlp_ReadNetSyncInfo(_socket.getDescriptor()); }

	synchronized int dlp_WriteNetSyncInfo(NetSyncInfo info) throws DlpException
		{ return calls.dlp_WriteNetSyncInfo(_socket.getDescriptor(), info); }

	public synchronized int dlp_ResetSyncFlags(int handle) throws DlpException
		{ return calls.dlp_ResetSyncFlags(_socket.getDescriptor(), handle); }

	synchronized int dlp_ResetSystem() throws DlpException
		{ return calls.dlp_ResetSystem(_socket.getDescriptor()); }

	synchronized byte[] dlp_CallApplication(Char4 creator, int type, int action, byte[] data, int[] retcode) throws DlpException
   		{ return calls.dlp_CallApplication(_socket.getDescriptor(), creator, type, action, data, retcode); }


	/**
	* Dlp shall only be constructed by the com.syncbuilder.sync.Link class.
	*/
	Dlp(Link link) 
		throws java.io.IOException
	{ 
		this._link = link;
		_socket = _link.getSocket();
		_impl = new com.syncbuilder.internal.sync.DlpImpl( _socket );
	}

	private Link _link;
	private Socket _socket;
	private com.syncbuilder.internal.sync.DlpImpl _impl;
}

