/* com/syncbuilder/sync/ServerSocket.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: ServerSocket.java,v 1.2 1998/09/20 18:25:40 christ Exp $ */

package com.syncbuilder.sync;

import java.io.*;

/** 
 * A representation of a Server socket.<br>
 * After its creation this socket sits there and waits for an incoming connection from a PalmOS device.
 * When a connection is established via accept(), a Link object is returned.
 *
 * @see Link
 * @see Dlp
 * @see Socket
 * @author Kenneth Albanowski, massive changes by Tilo Christ
 */
public class ServerSocket 
{
	/**
	* Construct a ServerSocket and bind it to a device. Then make it listen for incoming connections.
	*
	* @param device the name of the device. This is platform-dependent!
	*/
	public ServerSocket(String device) 
		throws IOException
	{	
		SocketImplFactory factory = Socket.getSocketFactory();

		_impl = (factory != null) ? (factory.createSocketImpl()) : (new com.syncbuilder.internal.sync.NativeSocketImpl());

		_impl.create();
		_impl.bind(device);
		_impl.listen(1);
	}


	/**
	* Wait for an incoming connection.<br>
	* This is a blocking call that returns as soon as a connection could
	* successfully be established. A Link object will be returned, through which
	* you can conduct all further communication.
	*/
    synchronized public Link accept() 
		throws IOException
    { 
		Socket socket = new Socket();
		_impl.accept(socket._impl);
		return new Link(socket); 
	}

	/**
	* Wait for an incoming connection with time-out.<br>
	* This is a blocking call that returns as soon as a connection could
	* successfully be established, or when the specified time-out has been reached.<br>
	* A Link object will be returned, through which you can conduct all 
	* further communication.
	*
	* @exception java.io.InterruptedIOException the timeout has expired
	*/
    synchronized public Link accept(int timeout) 
		throws IOException
    { 
		Socket socket = new Socket();
		
		_impl.accept(socket._impl, timeout);

		return new Link(socket); 
	}


	synchronized public void bind(String device) 
		throws IOException
	{ 
		_impl.bind(device); 
	}

    synchronized public void listen(int backlog) 
		throws IOException
    { 
		_impl.listen(backlog);
	}


	/**
	* Close the ServerSocket.<br>
	* This stops the ServerSocket from accepting clients.
	*/
	synchronized public void close() 
		throws IOException, DlpException 
	{ 
		_impl.close();
	}

	private final SocketImpl _impl;
}
