/* com/syncbuilder/util/PackUtils.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: PackUtils.java,v 1.2 1998/09/20 18:22:23 christ Exp $ */

package com.syncbuilder.util;


/**
* A collection of static helper-methods for packing/unpacking of the 
* native Records, AppInfos, SortInfos, etc.
*
* @author Tilo Christ
*/
final public class PackUtils
{
	/**
	* Extract a String from the data-array. The beginning of the
	* String within the array is specified by the offset. The end
	* of the string is automatically determined by looking for a 
	* 0x00-byte. 
	*
	* @param data the array of bytes from which the String is extracted
	* @param offset the beginning of the string within the array
	* @param newoff an Index, which will be set to the offset right behind
	*               the terminating 0x00-byte, or null if you are not interested
	*               in this information
	*
	* @return a Java.lang.String, with NO terminating 0x00-byte
	*/
	public static String getZeroTermString(byte[] data, int offset, Index newoff)
		throws java.lang.ArrayIndexOutOfBoundsException
	{
		int index2 = offset;

		while(data[index2++] != 0);

		String retval = new String(data, offset, (index2-offset) - 1);

		if (newoff != null)
		{
			newoff.setValue( index2 );
		}
		
		return retval;
	}


	/**
	* Place a String into an array of bytes. The string-data in the
	* array will be zero-terminated. The array therefore needs to be
	* 1 byte longer than the string. 
	*
	* @param data the array where the result will be placed
	* @param offset the place in the array where the result will be placed
	* @param the String. null is equivalent to an empty string
	*
	* @return the offset in the array right behind the terminating 0x00-byte
	*/
	public static int setZeroTermString(byte[] data, int offset, String text)
		throws java.lang.ArrayIndexOutOfBoundsException
	{		
		if (text != null)
		{
			byte[] textbytes = text.getBytes();
			
			System.arraycopy(textbytes, 0, data, offset, textbytes.length);
			offset += textbytes.length;
		}
		data[offset++] = (byte)0;

		return offset;
	}

	public static void setShort(byte[] data, int offset, int s)
		throws java.lang.ArrayIndexOutOfBoundsException
	{
		data[offset]   = (byte)((s >> 8) & 0xff);
		data[offset+1] = (byte)(s & 0xff);
	}

	public static int getShort(byte[] data, int offset)
		throws java.lang.ArrayIndexOutOfBoundsException
	{
		return ((data[offset] & 0xff) << 8) + (data[offset + 1] & 0xff);
	}

	public static int getLong(byte[] data, int offset)
		throws java.lang.ArrayIndexOutOfBoundsException
	{
		return ((data[offset] & 0xff) << 24) + ((data[offset + 1] & 0xff) << 16) + ((data[offset + 2] & 0xff) << 8) + (data[offset + 3] & 0xff);
	}

	public static void setLong(byte[] data, int offset, int s)
		throws java.lang.ArrayIndexOutOfBoundsException
	{
		data[offset]   = (byte)((s >> 24) & 0xff);
		data[offset+1]   = (byte)((s >> 16) & 0xff);
		data[offset+2]   = (byte)((s >> 8) & 0xff);
		data[offset+3] = (byte)(s & 0xff);
	}

	// Instantiation impossible
	private PackUtils()
	{
	}
}
