/* com/syncbuilder/util/TimeSlicer.java:
*
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: TimeSlicer.java,v 1.1 1998/09/25 06:36:01 christ Exp $ */

package com.syncbuilder.util;

import java.util.Vector;

/**
* This class simulates time-sliced multi-threading on a system
* that can only provide round-robin green threads.
* 
* @author Tilo Christ
*/
public class TimeSlicer
	implements Runnable
{
	public TimeSlicer(int base_priority)
	{
		if (base_priority + 2> Thread.MAX_PRIORITY)
			throw new IllegalArgumentException("base_priority too high!");
			
		this._base_priority = base_priority;
	}

	/**
	* The core of the scheduler. It works like this:
	* There is one master thread which runs with a priority of 
	* base_priority + 2 (this run()-method actually is the run()-
	* method of that thread). This master thread takes one of the
	* threads it is supposed to schedule, raises that thread's priority
	* to base_priority + 1, and then goes to sleep, thus giving the
	* thread with the raised priority a chance to run. When the
	* master thread wakes up, it puts the thread's priority back to
	* base_priority, like all the other threads, and picks the next
	* thread for execution, raises that thread's priority, ...
	*/
	public void run()
	{
		Thread this_thread = Thread.currentThread();
		
		while(this_thread == _master_thread)
		{
			// Wait until runnable threads become available			
			synchronized(this)
			{
				_threads = _new_threads;
				_new_threads = new Vector(50, 10);

				if (_threads.isEmpty())
				{
					try
					{
						wait();
					}
					catch (InterruptedException ie)
					{}
				}
			}

			for(int i=0; i < _threads.size(); i++)
			{
				Thread current_thread = (Thread)(_threads.elementAt(i));
				if (current_thread.isAlive())
				{
					try
					{
						current_thread.setPriority(_base_priority + 1);
					
						try
						{
							Thread.sleep(_time_slice);
						}
						catch(InterruptedException ie)
						{}
					
						current_thread.setPriority(_base_priority);

						// This thread will probably be runnable again during the next cycle
						// Therefore, we make sure we'll remember it :-)						
						_new_threads.addElement( current_thread ); 
					}
					catch (NullPointerException npe)
					{
						// NullPointerException means the thread is no longer alive.
					}
				}
			}
		}
	}

	public synchronized void start()
	{
		if (_master_thread == null)
		{
			_master_thread = new Thread(this);
			_master_thread.setPriority(_base_priority + 2);
			_master_thread.start();
		}
	}

	public synchronized void stop()
	{
		_master_thread = null;
	}

	public synchronized void addThread(Thread thread)
	{
		thread.setPriority(_base_priority);
		_new_threads.addElement(thread);
		
		notify();
	}
	
	private Thread _master_thread = null;
	private Vector _threads = new Vector(50, 10);   // The threads which are to be scheduled in the current cycle
	private Vector _new_threads = new Vector(50, 10); // The threads which are to be scheduled in the following cycle

	private int _base_priority;

	private int _time_slice = 50; // Size of the time-slice in ms.
}

